import numpy as np
from tuna_util import *


def antisymmetrise_integrals(ERI):

    """

    Antisymmetrises two-electron integrals.

    Args:   
        ERI (array): Electron repulsion integrals

    Returns:
        ERI_ansym (array): Antisymmetrised electron repulsion integrals

    """

    ERI_ansym = ERI - ERI.transpose(0,1,3,2)

    return ERI_ansym




def spin_block_molecular_orbitals(molecular_orbitals_alpha, molecular_orbitals_beta, epsilons):

    """

    Spin blocks alpha and beta molecular orbitals.

    Args:   
        molecular_orbitals_alpha (array): Alpha molecular orbitals in AO basis
        molecular_orbitals_beta (array): Beta molecular orbitals in AO basis
        epsilons (array): Orbital eigenvalues

    Returns:
        C_spin_block (array): Spin-blocked molecular orbitals in AO basis

    """

    C_spin_block = np.block([[molecular_orbitals_alpha, np.zeros_like(molecular_orbitals_beta)], [np.zeros_like(molecular_orbitals_alpha), molecular_orbitals_beta]])
    
    C_spin_block = C_spin_block[:, epsilons.argsort()] 

    return C_spin_block





def transform_ERI_AO_to_SO(ERI_AO, C_1, C_2):

    """

    Transforms electron repulsion integrals from the AO basis to the SO basis.

    Args:   
        ERI_AO (array): Electron repulsion integrals in AO basis
        C_1 (array): Molecular orbitals in AO basis
        C_2 (array): Molecular orbitals in AO basis

    Returns:
        ERI_SO (array): Electron repulsion integrals in SO basis

    """

    ERI_SO = np.einsum("mi,nj,mkln,ka,lb->ijab", C_1, C_2, ERI_AO, C_1, C_2, optimize=True)       

    return ERI_SO




def build_epsilons_tensor(epsilons_1, epsilons_2, o_1, o_2, v_1, v_2):

    """

    Builds inverse epsilon tensor with shape ijab.

    Args:   
        epsilons_1 (array): Orbital eigenvalues
        epsilons_2 (array): Orbital eigenvalues
        o_1 (slice): Occupied slice
        o_2 (slice): Occupied slice
        v_1 (slice): Virtual slice
        v_2 (slice): Virtual slice

    Returns:
        e_ijab (array): Inverse epsilons tensor with shape ijab

    """

    n = np.newaxis

    e_ijab = 1 / (epsilons_1[o_1, n, n, n] + epsilons_2[n, o_2, n, n] - epsilons_1[n, n, v_1, n] - epsilons_2[n, n, n, v_2])

    return e_ijab




def build_MP2_t_amplitudes(ERI_SO, e_ijab):

    """

    Build MP2 t amplitudes with shape ijab.

    Args:   
        ERI_SO (array): Electron-repulsion integrals in SO basis
        e_ijab (array): Inverse epsilons tensor shape ijab

    Returns:
        t_ijab (array): MP2 t amplitudes shape ijab

    """

    t_ijab = ERI_SO * e_ijab

    return t_ijab




def transform_matrix_AO_to_SO(M, C):

    """

    Transforms two-index tensor from AO basis to SO basis.

    Args:   
        M (array): Matrix in AO basis
        C (array): Molecular orbitals in AO basis

    Returns:
        M_SO (array): Matrix in SO basis

    """

    M_SO = C.T @ M @ C

    return M_SO





def transform_P_SO_to_AO(P_SO, C_spin_block, n_SO):

    """

    Transforms density matrix from SO basis to AO basis.

    Args:   
        P_SO (array): Density matrix in SO basis
        C_spin_block (array): Spin-blocked molecular orbitals in AO basis
        n_SO (int): Number of spin orbitals

    Returns:
        P_AO (array): Density matrix in AO basis
        P_alpha (array): Alpha density matrix in AO basis
        P_beta (array): Beta density matrix in AO basis

    """

    # Cuts spin-blocked molecular orbitals in two
    C_alpha = C_spin_block[:n_SO // 2, :]  
    C_beta = C_spin_block[n_SO // 2:, :] 

    # Transforms the alpha and beta parts of the total density matrix in the SO basis
    P_alpha = C_alpha @ P_SO @ C_alpha.T  
    P_beta = C_beta @ P_SO @ C_beta.T  

    P = P_alpha + P_beta

    return P, P_alpha, P_beta




def begin_spin_orbital_calculation(ERI_AO, SCF_output, n_occ):

    """

    Calculates key factors for spin orbital calculations.

    Args:   
        ERI_AO (array): Two electron integrals in AO basis
        SCF_output (Output): Output object
        n_occ (int): Number of occupied spin orbitals

    Returns:
        ERI_SO_anysm (array): Electron repulsion integrals in SO basis
        C_spin_block (array): Spin blocked molecular orbitals in AO basis
        epsilons_sorted (array): Sorted array of Fock matrix eigenvalues
        ERI_spin_block (array): Spin-blocked electron repulsion integrals in AO basis
        o (slice): Occupied spin orbitals
        v (slice): Virtual spin orbitals

    """
    
    # Defines occupied and virtual slices
    o = slice(None, n_occ)
    v = slice(n_occ, None)

    # Spin-blocks electron repulsion integrals
    ERI_spin_block = np.kron(np.eye(2), np.kron(np.eye(2), ERI_AO).T)

    epsilons_combined = SCF_output.epsilons_combined

    # Spin-blocks molecular orbitals and transforms electron repulsion integrals
    C_spin_block = spin_block_molecular_orbitals(SCF_output.molecular_orbitals_alpha, SCF_output.molecular_orbitals_beta, epsilons_combined)
    ERI_SO = transform_ERI_AO_to_SO(ERI_spin_block, C_spin_block, C_spin_block)

    # Antisymmetrise electron repulsion integrals
    ERI_SO_anysm = antisymmetrise_integrals(ERI_SO)

    # Sorts epsilons
    epsilons_sorted = np.sort(epsilons_combined)

    return ERI_SO_anysm, C_spin_block, epsilons_sorted, ERI_spin_block, o, v







def build_CIS_Hamiltonian(n_occ, n_virt, n_SO, epsilons, ERI_SO_ansym):
    
    """

    Contructs array of excitations and CIS Hamiltonian matrix in determinantal basis.

    Args:   
        n_occ (int): Number of occupied spin orbitals
        n_virt (int): Number of virtual spin orbitals
        n_SO (int): Number of spin orbitals
        epsilons (array): Fock matrix orbital eigenvalues
        ERI_SO_ansym (array): Antisymmetrised electron repulsion integrals in SO basis

    Returns:
        H_CIS (array): Shifted CIS Hamiltonian in determinantal basis
        excitations (array): List of possible excitations between occupied and virtual spin orbitals

    """
    
    excitations = []
    H_CIS = np.zeros((n_occ * n_virt, n_occ * n_virt))

    # Builds excitations list of every combination from occupied to virtual spin orbitals
    for i in range(n_occ):
        for a in range(n_occ, n_SO):

            excitations.append((i, a))

    # Builds shifted CIS Hamiltonian
    for p, (i, a) in enumerate(excitations):
        for q, (j, b) in enumerate(excitations):
        
            H_CIS[p, q] = (epsilons[a] - epsilons[i]) * (i == j) * (a == b) + ERI_SO_ansym[a, j, i, b]


    return H_CIS, np.array(excitations)





def calculate_weights_matrix(weights_of_interest, excitations, n_occ, n_virt):

    """

   Calculates the two index weight matrix for a state of interest.

    Args:   
        weights_of_interest (array): Weights array for state of interest
        excitations (array): List of possible excitations
        n_virt (int): Number of virtual spin orbitals

    Returns:
        b_ia (array): Two index weights matrix for state of interest

    """

    # Virtual indices shifted down to start counting the vertical indices from zero
    i_indices = excitations[:, 0]
    a_indices = excitations[:, 1] - n_occ

    b_ia = np.zeros((n_occ, n_virt))

    b_ia[i_indices, a_indices] = weights_of_interest

    return b_ia





def calculate_CIS_density_matrix(n_SO, n_occ, C_spin_block, o, v, b_ia):

    """

    Calculates the CIS density matrix for a state of interest.

    Args:   
        n_SO (int): Number of spin orbitals
        n_occ (int): Number of occupied spin orbitals
        C_spin_block (array): Spin-blocked molecular orbitals in AO basis
        o (slice): Occupied orbitals slice
        v (slice): Virtual orbitals slice
        b_ia (array): Weights matrix for state of interest

    Returns:
        P (array): Density matrix in AO basis
        P_alpha (array): Density matrix for alpha orbitals in AO basis
        P_beta (array): Density matrix for beta orbitals in AO basis

    """
    
    # Equation from Foresman 1992 - this is the transition density, delta P_CIS
    DP_ij = -1 * np.einsum("ia,ja->ij", b_ia, b_ia, optimize=True)
    DP_ab = np.einsum("ia,ib->ab", b_ia, b_ia, optimize=True)

    # Initialises HF density matrix in SO basis
    P_SO = np.zeros((n_SO, n_SO))
    P_SO[o, o] = np.identity(n_occ)

    # Adds on transition density blocks to form unrelaxed density matrix in SO basis
    P_SO[v, v] += DP_ab
    P_SO[o, o] += DP_ij

    P, P_alpha, P_beta = transform_P_SO_to_AO(P_SO, C_spin_block, n_SO)

    return P, P_alpha, P_beta






def calculate_transition_dipoles(D, weights, excitations, C_spin_block):

    """

    Calculates the transition dipoles (measure of intensity) from a CIS calculation.

    Args:   
        D (array): Dipole integrals
        weights (array): Weights of each excitation (determinant) from CIS calculation
        excitations (array): List of possible excitations
        C_spin_block (array): Spin-blocked molecular orbitals in AO basis

    Returns:
        transition_dipoles (array): Transition dipoles

    """
    
    D_spin_block = np.kron(np.eye(2), D)
    D_SO = transform_matrix_AO_to_SO(D_spin_block, C_spin_block)

    # Left column is occupied, right column is virtual
    i_indices = excitations[:, 0]
    a_indices = excitations[:, 1]

    transition_dipoles = weights.T @ D_SO[a_indices, i_indices]

    return transition_dipoles





def calculate_oscillator_strengths(transition_dipoles, excitation_energies):

    """

    Calculates the oscillator strengths of all states at once.

    Args:   
        transition_dipoles (array): Transition dipoles for each state
        excitation_energies (array): Excitation energies for each state

    Returns:
        oscillator_strengths (array): Oscillator strengths for states

    """
    
    oscillator_strengths = (2 / 3) * excitation_energies * transition_dipoles ** 2

    return oscillator_strengths






def label_spin_orbital(index, calculation):

    """

    Labels and formats the spin orbitals for logging.

    Args:   
        index (int): Index of spin orbital
        calculation (Calculation): Calculation object

    Returns:
        labelled_SO (string): Formatted spin orbital label

    """
    
    reduced_index = index // 2 + 1

    labelled_SO = str(reduced_index)

    # Adds an "a" for alpha and "b" for beta for UHF
    if calculation.reference == "UHF":
        
        labelled_SO += "a" if index % 2 == 0 else "b"
    
    return labelled_SO






def print_excited_state_information(excitation_energies, SCF_output, contributions, weights, excitations_formatted, calculation, silent=False):
  
    """

    Prints information about each excited state.

    Args:   
        excitation_energies (array): Excitation energies
        SCF_output (Output): Output from SCF calculation
        contributions (array): Percent contributions to each state
        weights (array): Weights for each state
        excitations_formatted (array): Formatted excitations
        calculation (Calculation): Calculation object
        silent (bool, optional): Should output be silenced

    Returns:
        None: Nothing is returned

    """

    for state in range(len(excitation_energies)):

        if state < calculation.n_states:
            
            # Prints excitation energy and energy of each state
            log(f"\n\n  ~~~~~ State {state + 1} ~~~~~   ", calculation, 2, silent=silent)

            log(f"\n  Excitation energy:   {excitation_energies[state]:13.10f}", calculation, 2, silent=silent)
            log(f"  Energy of state:     {(excitation_energies[state] + SCF_output.energy):13.10f}\n", calculation, 2, silent=silent)
            
            # Prints contributions if they are above the threshold
            for i, excitation in enumerate(excitations_formatted):

                if contributions[i, state] > calculation.CIS_contribution_threshold:

                    log(f"    {excitation[0]} -> {excitation[1]}  :  {contributions[i, state]:6.2f} %     ({weights[i, state]:9.6f})", calculation, 2, silent=silent)





def print_CIS_absorption_spectrum(molecule, excitation_energies_eV, calculation, frequencies_per_cm, wavelengths_nm, transition_dipoles, oscillator_strengths, silent=False):
    
    """

    Prints CIS absorption spectrum information.

    Args:   
        molecule (Molecule): Molecule object
        excitation_energies_eV (array): Excitation energies in eV
        calculation (Calculation): Calculation object
        frequencies_per_cm (array): Frequencies in per cm
        wavelengths_nm (array): Wavelengths in nm
        transition_dipoles (array): Transition dipoles
        oscillator_strengths (array): Oscillator strengths
        silent (bool, optional): Should output be silenced

    Returns:
        None: Nothing is returned

    """
   
    log(f"\n\n Transition dipole moment origin is the centre of mass, {bohr_to_angstrom(molecule.centre_of_mass):.4f} angstroms from the first atom.", calculation, 1, silent=silent)
    
    log("\n ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~", calculation, 1, silent=silent)
    log("                                          CIS Absorption Spectrum", calculation, 1, silent=silent, colour="white")
    log(" ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~", calculation, 1, silent=silent)
    log("   State     Energy (eV)     Freq. (per cm)     Wavelength (nm)     Osc. Strength     Transition Dipole", calculation, 1, silent=silent)
    log(" ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~", calculation, 1, silent=silent)

    for state in range(len(excitation_energies_eV)):

        if state < calculation.n_states:

            log(f"    {(state + 1):2}         {excitation_energies_eV[state]:7.4f}          {frequencies_per_cm[state]:8.1f}            {wavelengths_nm[state]:5.1f}              {oscillator_strengths[state]:.6f}          {transition_dipoles[state]:10.7f}", calculation, 1, silent=silent)

    log(" ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~", calculation, 1, silent=silent)





def calculate_CIS_doubles_correction(excitation_energy, epsilons, root, ERI_SO_ansym, o, v, b_ia, calculation, silent=False):

    """

    Calculates CIS(D) correction to CIS excitation energy of one state.

    Args:   
        excitation_energy (float): Excitation energy of state of interest
        epsilons (array): Fock matrix orbital eigenvalues
        root (int): State of interest
        ERI_SO_ansym (array): Electron repulsion integrals in SO basis, antisymmetrised
        o (slice): Occupied orbital slice
        v (slice): Virtual orbital slice
        b_ia (array): Two-index weights matrix for state of interest
        calculation (Calculation): Calculation object
        silent (bool, optional): Should output be silenced

    Returns:
        E_CIS_D (float): CIS(D) correction to CIS excitation energy

    """

    # Equations taken from Head-Gordon paper
    log("\n ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~", calculation, 1, silent=silent)
    log("          CIS(D) Perturbative Correction", calculation, 1, silent=silent, colour="white")
    log(" ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~", calculation, 1, silent=silent)

    log(f"  Applying doubles correction to state {root + 1} only.", calculation, 1, silent=silent)
   
    log(f"\n  Building MP2 amplitudes...            ", calculation, 1, silent=silent, end=""); sys.stdout.flush()
    
    # Builds and inverts inverse epsilons tensor, to upright e_ijab_inv
    e_ijab_inv = 1 / build_epsilons_tensor(epsilons, epsilons, o, o, v, v)
    e_ijab_inv_minus_w = 1 / (e_ijab_inv + excitation_energy)

    t_ijab = build_MP2_t_amplitudes(ERI_SO_ansym[o, o, v, v], 1 / e_ijab_inv)

    log(f"  [Done]", calculation, 1, silent=silent)

    log(f"\n  Calculating direct contribution...  ", calculation, 1, silent=silent, end=""); sys.stdout.flush()
    
    u_1 = np.einsum("abcj,ic->ijab", ERI_SO_ansym[v, v, v, o], b_ia, optimize=True)
    u_2 = np.einsum("abci,jc->ijab", ERI_SO_ansym[v, v, v, o], b_ia, optimize=True)
    u_3 = np.einsum("kaij,kb->ijab", ERI_SO_ansym[o, v, o, o], b_ia, optimize=True)
    u_4 = np.einsum("kbij,ka->ijab", ERI_SO_ansym[o, v, o, o], b_ia, optimize=True)

    u_ijab = u_1 - u_2 + u_3 - u_4

    log(f"    [Done]", calculation, 1, silent=silent)

    log(f"  Calculating indirect contribution...  ", calculation, 1, silent=silent, end=""); sys.stdout.flush()
    
    v_1 = (1 / 2) * np.einsum("jkbc,ib,jkca->ia", ERI_SO_ansym[o, o, v, v], b_ia, t_ijab, optimize=True)
    v_2 = (1 / 2) * np.einsum("jkbc,ja,ikcb->ia", ERI_SO_ansym[o, o, v, v], b_ia, t_ijab, optimize=True)
    v_3 = np.einsum("jkbc,jb,ikac->ia", ERI_SO_ansym[o, o, v, v], b_ia, t_ijab, optimize=True)

    v_ia = v_1 + v_2 + v_3

    log(f"  [Done]", calculation, 1, silent=silent)

    log(f"\n  Calculating CIS(D) correction...      ", calculation, 1, silent=silent, end=""); sys.stdout.flush()
    
    E_CIS_D = (1 / 4) * np.einsum("ijab,ijab,ijab->", u_ijab, u_ijab, e_ijab_inv_minus_w, optimize=True) + np.einsum("ia,ia->", b_ia, v_ia, optimize=True)

    log(f"  [Done]", calculation, 1, silent=silent)
    
    # Correction energy in eV prints if P used
    log(f"\n  Excitation energy from CIS:      {excitation_energy:13.10f}", calculation, 1, silent=silent)
    log(f"  Correction energy from CIS(D):   {E_CIS_D:13.10f}", calculation, 1, silent=silent)
    log(f"  Correction energy (eV):          {(E_CIS_D * constants.eV_in_hartree):13.10f}", calculation, 3, silent=silent)
    log(f"\n  Excitation energy from CIS(D):   {(E_CIS_D + excitation_energy):13.10f}", calculation, 1, silent=silent)
    log(" ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~", calculation, 1, silent=silent)
  
    return E_CIS_D





def run_CIS(ERI_AO, n_occ, n_virt, n_SO, calculation, SCF_output, molecule, silent=False):

    """

    Begins a configuration interaction singles calculation.

    Args:   
        ERI_AO (array): Electron repulsion integrals in AO basis
        n_occ (int): Number of occupied spin orbitals
        n_virt (int): Number of virtual spin orbitals
        n_SO (int): Number of spin orbitals
        calculation (Calculation): Calculation object
        SCF_output (Output): SCF output object
        molecule (Molecule): Molecule object
        silent (bool, optional): Should output be silenced

    Returns:
        E_CIS (float): Energy of state of interest
        E_transition (float): Transition energy to state of interest
        P_CIS (array): Density matrix of state of interest in AO basis      
        P_CIS_a (array): Density matrix of alpha orbitals of state of interest in AO basis
        P_CIS_b (array): Density matrix of beta orbitals of state of interest in AO basis

    """
    
    # Converting into computer counting from human counting
    root = calculation.root - 1

    log("\n ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~", calculation, 1, silent=silent)
    log("         Configuration Interaction Singles", calculation, 1, silent=silent, colour="white")
    log(" ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n", calculation, 1, silent=silent)
    
    log("  Transforming two-electron integrals... ", calculation, 1, silent=silent, end=""); sys.stdout.flush()

    ERI_SO_ansym, C_spin_block, epsilons_sorted, _, o, v = begin_spin_orbital_calculation(ERI_AO, SCF_output, n_occ)

    log(" [Done]", calculation, 1, silent=silent)

    log("\n  Building CIS Hamiltonian...             ", calculation, 1, silent=silent, end=""); sys.stdout.flush()

    H_CIS, excitations = build_CIS_Hamiltonian(n_occ, n_virt, n_SO, epsilons_sorted, ERI_SO_ansym)

    log("[Done]", calculation, 1, silent=silent)

    log("  Diagonalising CIS Hamiltonian...        ", calculation, 1, silent=silent, end=""); sys.stdout.flush()

    excitation_energies, weights = np.linalg.eigh(H_CIS)

    percent_contributions = np.round(weights ** 2 * 100)

    frequencies_per_cm = excitation_energies * constants.per_cm_in_hartree
    wavelengths_nm = 1e7 / frequencies_per_cm
    excitation_energies_eV = constants.eV_in_hartree * excitation_energies

    log("[Done]", calculation, 1, silent=silent)

    log("  Calculating transition dipoles...       ", calculation, 1, silent=silent, end=""); sys.stdout.flush()

    transition_dipoles = calculate_transition_dipoles(SCF_output.D, weights, excitations, C_spin_block)
    oscillator_strengths = calculate_oscillator_strengths(transition_dipoles, excitation_energies)

    log("[Done]", calculation, 1, silent=silent)

    log(f"  Constructing density matrix...          ", calculation, 1, silent=silent, end=""); sys.stdout.flush()

    weights_of_interest = weights[:, root]

    # Calculates weight matrix for state of interest
    b_ia = calculate_weights_matrix(weights_of_interest, excitations, n_occ, n_virt)
    P_CIS, P_CIS_a, P_CIS_b = calculate_CIS_density_matrix(n_SO, n_occ, C_spin_block, o, v, b_ia)

    log("[Done]", calculation, 1, silent=silent)

    log("\n  Printing excited state information...  ", calculation, 2, silent=silent)

    excitations_formatted = [(label_spin_orbital(i, calculation), label_spin_orbital(a, calculation)) for i, a in excitations]

    print_excited_state_information(excitation_energies, SCF_output, percent_contributions, weights, excitations_formatted, calculation, silent=silent)

    print_CIS_absorption_spectrum(molecule, excitation_energies_eV, calculation, frequencies_per_cm, wavelengths_nm, transition_dipoles, oscillator_strengths, silent=silent)

    # Energy of transition of state of interest
    E_transition = excitation_energies[root]

    # Optionally applies doubles correction to transition energy for a specified state
    if "[D]" in calculation.method:
            
        E_CIS_D = calculate_CIS_doubles_correction(E_transition, epsilons_sorted, root, ERI_SO_ansym, o, v, b_ia, calculation, silent=silent)

        E_transition += E_CIS_D


    E_CIS = SCF_output.energy + E_transition


    return E_CIS, E_transition, P_CIS, P_CIS_a, P_CIS_b


